/* Copyright (C) 2005 Theodore A. Phelps */
/* Copyright (C) 2005 Free Software Foundation, Inc.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.  */

#if !defined(lint)
static const char cvsid[] = "@(#)$Id: speedwatch.c,v 1.109 2006/02/11 11:00:29 phelps Exp $";
#endif

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#ifdef HAVE_STDIO_H
#include <stdio.h>
#endif
#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif
#ifdef HAVE_STDINT_H
#include <stdint.h>
#endif
#ifdef HAVE_STRING_H
#include <string.h>
#endif
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_PATHS_H
#include <paths.h>
#endif
#ifdef HAVE_SYS_TYPES_H
#include <sys/types.h>
#endif
#ifdef HAVE_SYS_SOCKET_H
#include <sys/socket.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif
#ifdef HAVE_SYS_TIME_H
#include <sys/time.h>
#endif
#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif
#ifdef HAVE_SYS_IOCTL_H
#include <sys/ioctl.h>
#endif
#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYSLOG_H
#define SYSLOG_NAMES 1
#include <syslog.h>
#endif
#ifdef HAVE_SIGNAL_H
#include <signal.h>
#endif
#ifdef HAVE_STDARG_H
#include <stdarg.h>
#endif
#ifdef HAVE_ALLOCA_H
#include <alloca.h>
#endif
#ifdef HAVE_ERRNO_H
#include <errno.h>
#endif
#include <linux/atm.h>
#include <linux/atmdev.h>
#include <linux/atmbr2684.h>

/*
 * Although the speedtouch driver and firmware are reasonably stable,
 * the ATM link will occasionally drop out, killing the network
 * connection.  Restarting usually involves removing the driver module
 * and re-inserting it to reset the state.
 *
 * Speedwatch should ensure that the speedtouch modem is restarted
 * when the ATM link drops out.  It will load the speedtch module and,
 * after a judicious pause, start pppd using the nodetach option to
 * keep it in the foreground.  By default, if pppd dies with an exit
 * code of 15 ("the link was terminated because te peer is not
 * responding to echo requests") then the speedtch module will be
 * unloaded and, after a few seconds, re-loaded, restarting the loop.
 *
 * If pppd dies with any other signal, the speedtch module will be
 * removed and this program will exit with pppd's exit code.
 *
 * If speedwatch receives a SIGINT or SIGTERM signal, it will unload
 * the speedtch module and quit with a 5 exit code, which corresponds
 * to pppd's exit code 5 ("pppd terminated because it was sent a
 * SIGINT, SIGTERM or SIGHUP).
 *
 * If speedwatch receives a SIGHUP signal, it will kill pppd and
 * behave as if it had died with an exit code of 15.
 */

/* pppd exit codes from pppd.h */
#define EXIT_OK 0
#define EXIT_FATAL_ERROR 1
#define EXIT_OPTION_ERROR 2
#define EXIT_NOT_ROOT 3
#define EXIT_NO_KERNEL_SUPPORT 4
#define EXIT_USER_REQUEST 5
#define EXIT_LOCK_FAILED 6
#define EXIT_OPEN_FAILED 7
#define EXIT_CONNECT_FAILED 8
#define EXIT_PTYCMD_FAILED 9
#define EXIT_NEGOTIATION_FAILED 10
#define EXIT_PEER_AUTH_FAILED 11
#define EXIT_IDLE_TIMEOUT 12
#define EXIT_CONNECT_TIME 13
#define EXIT_CALLBACK 14
#define EXIT_PEER_DEAD 15
#define EXIT_HANGUP 16
#define EXIT_LOOPBACK 17
#define EXIT_INIT_FAILED 18
#define EXIT_AUTH_TOPEER_FAILED 19
#define EXIT_TRAFFIC_LIMIT 20
#define EXIT_CNID_AUTH_FAILED 21

/* Uncommon system call */
extern long delete_module(const char *name, unsigned int flags);

/* The size of the buffer used to read from proc files */
#define BUFFER_SIZE 1024

/* The number of us to wait after the line is up before starting pppd */
#define LINE_UP_SETTLE_PAUSE_USEC (2 * 1000 * 1000)

/* The number of us to wait after unloading the modules before trying
 * to load them again. */
#define UNLOAD_SETTLE_PAUSE_USEC (15 * 1000 * 1000)

/* The number of us to wait when polling the modem status */
#define SHORT_PAUSE_USEC (500 * 1000)


/* The number of us to wait before deciding that the modem simply
 * doesn't exist */
#define MODEM_INIT_TIMEOUT (30 * 1000 * 1000)

/* The number of us to wait before deciding that the ATM negotiation
 * isn't going to happen */
#define MODEM_LINEUP_TIMEOUT (5 * 60 * 1000 * 1000)

/* When reading the modem status file, the number of times that we
 * fail to interpret it before we give up */
#define MAX_MALFORMED 5

/* Maximum number of authentication failures without a successful
 * connection before we give up */
#define MAX_AUTH_ERRORS 3

/* Arguments for modprobe */
#define MODPROBE_ARGS "-s"

/* Paths */
#define PATH_PIDFILE "/var/run/speedwatch.pid"
#define PATH_PPPD "/usr/sbin/pppd"
#define PATH_PROC "/proc"
#define PATH_PROC_MODULES PATH_PROC "/modules"
#define PATH_PROC_MODPROBE PATH_PROC "/sys/kernel/modprobe"
#define PATH_PROC_ATM PATH_PROC "/net/atm"
#define PATH_PROC_DEVICES PATH_PROC_ATM "/devices"
#define PATH_IFCONFIG "/sbin/ifconfig"

#define SET_STATE(var, state) \
    do { \
        xlog(LOG_DEBUG, "changing to state " #state); \
        var = state; \
    } while (0)


/* The FSM states */
typedef enum
{
    /* Initial state: module and firmware not loaded, pppd not running */
    FSM_START,

    /* Module loaded.  Waiting for device status file to appear */
    FSM_FIND_STATUS,

    /* Module loaded.  Waiting for firmware to load */
    FSM_FIRMWARE,

    /* Bringing up the ethernet bridge */
    FSM_BRIDGE,

    /* A brief pause to let things settle */
    FSM_PAUSE,

    /* Daemonize */
    FSM_DAEMONIZE,

    /* Start pppd */
    FSM_PPPD,

    /* Wait for pppd to die */
    FSM_RUNNING,

    /* Something has gone wrong.  Kill pppd, unload speedtch and wait
     * for things to settle before going back to the start state */
    FSM_ERROR
} fsm_state_t;

/* The speedtouch modem status */
typedef enum
{
    /* The status file can't be found.  Either the speedtch module
     * hasn't been loaded or no modem was detected. */
    STATUS_NO_MODEM,

    /* Modem found, firmware loaded, line is down */
    STATUS_LINE_DOWN,

    /* Modem found, firmware loaded, line status unknown */
    STATUS_LINE_UNKNOWN,

    /* Modem found, firmware loaded, line is up */
    STATUS_LINE_UP,
} status_t;

/* Internal error codes and their textual format */
char *int_err_strings[] =
{
    "No error",
#define EPARSE -1
    "Malformed input",
#define EBADEXIT -2
    "Exit code indicates failure"
#define MIN_ERROR -3
};

/* Map status names to strings */
char *status_names[] =
{
    "modem not found",
    "line is down",
    "line status unknown",
    "line is up"
};

struct dead_child
{
    /* The pid of the deceased */
    pid_t pid;

    /* The child's exit code */
    int code;

    /* The child's exit status */
    int status;
};


/* The name of the executable, stripped of leading path */
static char *progname;

/* The path to the modprobe executable */
static char *path_modprobe;

/* The list of modules to unload/load to reset the modem */
static char *mod_list = NULL;

/* The path to the modem status file */
static char path_modem_status[BUFFER_SIZE];

/* How much noise to make */
static int verbosity = LOG_WARNING;

/* Zero if we've daemonized */
static int in_foreground = 1;

/* The ATM interface specified on the command-line */
static int atm_itf = -1;

/* The ATM address */
struct sockaddr_atmpvc atm_addr;

/* The file descriptor of the ATM back-end socket */
static int atm_fd = -1;

/* The MAC or ESI (end-system interface) address */
static const char *mac_addr;

/* Where to record our pid */
static char *pid_file = PATH_PIDFILE;

/* The pid of the pppd process, or  -1 if it isn't running */
static int pppd_pid = -1;

/* Non-zero if we've received a SIGHUP */
static int do_restart = 0;

/* Non-zero if we've received a SIGINT or SIGTERM */
static int do_shutdown = 0;

/* Whitespace characters */
static char *ws = " \t\n";

/* If the pid of the morgue is not (pid_t)-1 then it represents an
 * unprocessed dead child */
struct dead_child morgue;

/* The file descriptor of the bridge control socket */
int brigectl_fd;

#define OPTIONS "a:e:Ff:hi:m:p:u:v:V"
static void
usage()
{
    fprintf(stderr,
            "usage: %s [OPTION...] pppd-arg...\n"
            "    -a [itf.]vpi.vci\tUse PPPoE with given parameters\n"
            "    -e llc|vcmux\tEnable virtual circuit encapsulation for PPPoE [llc]\n"
            "    -u mod-list\t\tComma-separated list of modules\n"
            "    -f facility\t\tWrite syslog messages to facility [daemon]\n"
            "    -i num\t\tUse the ATM device on interface num\n"
            "    -m addr\t\tUse the ATM device with mac address addr\n"
            "    -p file\t\tRecord process id in file [" PATH_PIDFILE "]\n"
            "    -F\t\t\tRun in the foreground\n"
            "    -v num\t\tSet the verbosity level [warning]\n"
            "    -h\t\t\tPrint this help message\n"
            "    -V\t\t\tPrint version information and exit\n",
            progname);
}

/* Write a message to syslog or stderr, as appropriate.  Note that
 * xlog will append a linefeed if necessary, so the format string
 * should not end with one. */
static void xlog(int priority, char *format, ...)
{
    va_list ap;

    /* Bail if the log level is too high */
    if (verbosity < priority) {
        return;
    }

    va_start(ap, format);
    
    /* If we're running in the foreground then print to stderr.
     * Otherwise write to syslog. */
    if (in_foreground) {
        fprintf(stderr, "%s: ", progname);
        vfprintf(stderr, format, ap);
        fputc('\n', stderr);
    } else {
        vsyslog(priority, format, ap);
    }

    va_end(ap);
}

/* Convert an error code into a string */
static const char *xstrerror(int err)
{
    if (MIN_ERROR < err && err < 0) {
        return int_err_strings[-err];
    } else {
        return strerror(err);
    }
}

/* Get the current time, or die trying */
static void xgettimeofday(struct timeval *tv)
{
    if (gettimeofday(tv, NULL) < 0) {
        xlog(LOG_ERR, "error: gettimeofday failed: %s", xstrerror(errno));
        exit(1);
    }
}

/* Handler for SIGCHLD */
static void handle_sigchld(int signum, siginfo_t *siginfo, void *ignored)
{
    /* Sanity check */
    if (morgue.pid != (pid_t)-1) {
        xlog(LOG_ERR, "internal error: multiple children in the morgue");
        abort();
    }

    /* Record the process id and exit information of the child */
    morgue.pid = siginfo->si_pid;
    morgue.code = siginfo->si_code;
    morgue.status = siginfo->si_status;
}

/* Handler for SIGHUP */
static void handle_sighup(int signum, siginfo_t *siginfo, void *ignored)
{
    xlog(LOG_INFO, "received SIGHUP: restarting");
    do_restart = 1;
}

/* Handler for SIGINT and SIGTERM */
static void handle_sigterm(int signum, siginfo_t *siginfo, void *ignored)
{
    xlog(LOG_INFO, "shutting down on signal %d", signum);
    do_shutdown = 1;
}

/* Clean up shop so that we can be a good daemon */
static void daemonize()
{
    pid_t pid;
    long fd_count;
    int fd;

    /* Fork a child process */
    if ((pid = fork()) == (pid_t)-1) {
        xlog(LOG_ERR, "error: unable to spawn child process: %s", xstrerror(errno));
        exit(-1);
    }

    /* If we're the parent then exit */
    if (pid != 0) {
        exit(0);
    }

    /* Create a new session and lose the controlling TTY */
    if (setsid() == (pid_t)-1) {
        xlog(LOG_ERR, "error: unable to set the session id: %s", xstrerror(errno));
        exit(1);
    }

    /* Fork again so that we can never acquire a controlling TTY */
    if ((pid = fork()) == (pid_t)-1) {
        xlog(LOG_ERR, "error: unable to spawn child process: %s", xstrerror(errno));
        exit(1);
    }

    /* The parent must die */
    if (pid != 0) {
        exit(1);
    }

    /* Move to the root directory */
    if (chdir("/") < 0) {
        xlog(LOG_ERR, "error: unable to change to root directory: %s", xstrerror(errno));
        exit(1);
    }

    /* Set the umask to zero */
    umask(0);

    /* Determine how many file descriptors we're dealing with */
    fd_count = sysconf(_SC_OPEN_MAX);

    /* Make sure that all open files are closed */
    for (fd = 0; fd < fd_count; fd++) {
        close(fd);
    }

    /* Replace stdin, stdout and stderr with /dev/null */
    if (open(_PATH_DEVNULL, O_RDONLY) < 0 ||
        (fd = open(_PATH_DEVNULL, O_WRONLY)) < 0 ||
        dup2(fd, 2) < 0) {
        xlog(LOG_ERR, "error: unable to set up stdin, stdout and stderr: %s",
             xstrerror(errno));
        exit(1);
    }
}

/* Check the modem line status.  Returns 0 to indicate that all is
 * well, -1 to say that it doesn't understand the status file and
 * errno to indicate any other error. */
static int modem_get_status(char *path, status_t *status_out)
{
    FILE *file;
    char buffer[BUFFER_SIZE];
    char *point;
    char *word;
    int i, err;

    /* Open the stats file */
    if ((file = fopen(path, "r")) == NULL) {
        if (errno == ENOENT) {
            *status_out = STATUS_NO_MODEM;
            return 0;
        } else {
            return errno;
        }
    }

    /* Read the first four lines, discarding all but the last */
    for (i = 0; i < 4; i++) {
        if ((fgets(buffer, BUFFER_SIZE, file)) == NULL) {
            if (ferror(file)) {
                err = errno;
                fclose(file);
                return err;
            }

            /* Premature end of file */
            fclose(file);
            return EPARSE;
        }
    }

    /* We're done with the file, so close it */
    if (fclose(file) < 0) {
        return errno;
    }

    /* Make sure the first word is "Line" */
    point = buffer;
    word = strsep(&point, ws);
    if (strcmp(word, "Line") != 0) {
        return EPARSE;
    }

    /* According to usbatm.c, the next word(s) should be
     * "up", "down" or "state unknown" */
    word = strsep(&point, ws);
    if (strcmp(word, "up") == 0) {
        *status_out = STATUS_LINE_UP;
        return 0;
    }

    if (strcmp(word, "down") == 0) {
        *status_out = STATUS_LINE_DOWN;
        return 0;
    }

    if (strcmp(word, "state") == 0) {
        /* Assume that the next word is "unknown" */
        *status_out = STATUS_LINE_UNKNOWN;
        return 0;
    }

    /* Can't parse source file */
    return EPARSE;
}


/* Get the path to modprobe from /proc */
static int find_modprobe()
{
    int err;
    char buffer[BUFFER_SIZE];
    FILE *file;
    char *point;

    /* Open /proc/sys/kernel/modprobe */
    if ((file = fopen(PATH_PROC_MODPROBE, "r")) == NULL) {
        return errno;
    }
    
    /* Read the one line in it */
    if (fgets(buffer, BUFFER_SIZE, file) == NULL) {
        err = errno;
        fclose(file);
        return err;
    }

    /* Kill the linefeed */
    if ((point = strchr(buffer, '\n')) != NULL) {
        *point = 0;
    }

    /* Duplicate the string */
    if ((path_modprobe = strdup(buffer)) == NULL) {
        err = errno;
        fclose(file);
        return err;
    }

    /* Close it */
    if (fclose(file) < 0) {
        return errno;
    }

    return 0;
}

/* Canonicalize a module name */
static void module_canonical_name(const char *name,
                                  char *canon_out,
                                  size_t canon_size)
{
    const char *in = name;
    char *out = canon_out;

    while (*in != 0 && out < canon_out + canon_size - 1) {
        if (*in == '-') {
            *out = '_';
        } else {
            *out = *in;
        }

        in++;
        out++;
    }

    *out = 0;
}

/* See if a module is loaded */
static int module_is_loaded(const char *name, int *found_out)
{
    FILE *file;
    char buffer[BUFFER_SIZE];
    char *module_name;
    char *point;
    int err = 0;

    /* Assume not found */
    *found_out = 0;

    /* Open /proc/modules */
    if ((file = fopen(PATH_PROC_MODULES, "r")) == NULL) {
        return errno;
    }

    for (;;) {
        /* Read the next line */
        if (fgets(buffer, BUFFER_SIZE, file) == NULL) {
            if (ferror(file)) {
                err = errno;
            }

            break;
        }

        /* Extract the module name */
        point = buffer;
        module_name = strsep(&point, ws);
        if (strcmp(module_name, name) == 0) {
            *found_out = 1;
            break;
        }
    }

    fclose(file);
    return err;
}

/* Load a module */
static int module_load(const char *name)
{
    int err;
    int found;
    int status;
    pid_t pid;
    char canon[BUFFER_SIZE];

    /* Canonicalize the module name */
    module_canonical_name(name, canon, BUFFER_SIZE);

    /* See if the module is already present */
    if ((err = module_is_loaded(canon, &found)) != 0) {
        return err;
    }

    /* Bail if it is */
    if (found) {
        return 0;
    }

    /* Make a note */
    xlog(LOG_DEBUG, "loading module %s", canon);

    /* Spawn yet another process */
    if ((pid = fork()) == (pid_t)-1) {
        return errno;
    }

    if (pid == 0) {
        /* We're the child; execute modprobe. */
        execl(path_modprobe, "modprobe",
              MODPROBE_ARGS,
              canon, NULL);

        xlog(LOG_ERR, "error: unable to exec %s: %s",
             path_modprobe, xstrerror(errno));
        exit(1);
    }

    /* We're the parent.  Wait for the child to die. */
    if (waitpid(pid, &status, 0) == (pid_t)-1) {
        return errno;
    }

    /* Check for a happy exit code */
    if (WIFEXITED(status) && WEXITSTATUS(status) == 0) {
        return 0;
    }

    /* Modprobe failed, but we don't know how. */
    return EBADEXIT;
}

/* Load each of the modules in a comma-separated list */
static int modules_load(const char *list)
{
    char buffer[BUFFER_SIZE];
    const char *mark = list;
    const char *point;
    int err;

    for (point = list; *point != 0; point++) {
        if (*point == ',') {
            memcpy(buffer, mark, point - mark);
            buffer[point - mark] = 0;
            if ((err = module_load(buffer)) != 0) {
                return err;
            }

            mark = point + 1;
        }
    }

    return module_load(mark);
}

/* Unload a module */
static int module_remove(const char *name)
{
    int err;
    int found;
    char canon[BUFFER_SIZE];

    /* Canonicalize the module name */
    module_canonical_name(name, canon, BUFFER_SIZE);

    /* Make sure the module is present */
    if ((err = module_is_loaded(canon, &found)) != 0) {
        return err;
    }

    /* If not, we're done */
    if (!found) {
        return 0;
    }

    /* Make a note */
    xlog(LOG_DEBUG, "removing module %s", canon);

    /* Unload the module */
    if (delete_module(canon, 0) < 0) {
        return errno;
    }

    return 0;
}

/* Removes each of the modules in a comma-separated list */
static int modules_remove(const char *list)
{
    char buffer[BUFFER_SIZE];
    const char *mark = list + strlen(list);
    const char *point;
    int err;

    for (point = mark; point >= list; point--) {
        if (*point == ',') {
            memcpy(buffer, point + 1, mark - point - 1);
            buffer[mark - point - 1] = 0;
            if ((err = module_remove(buffer)) != 0) {
                return err;
            }

            mark = point;
        }
    }

    memcpy(buffer, list, mark - list);
    buffer[mark - list] = 0;
    return module_remove(buffer);
}



/* Invoke ifconfig */
static int ifconfig(char *iface, char *cmd)
{
    pid_t pid;
    int status;

    /* Spawn a child process */
    if ((pid = fork()) == (pid_t)-1) {
        return errno;
    }

    if (pid == 0) {
        /* We're the child; execute ifconfig */
        execl(PATH_IFCONFIG, PATH_IFCONFIG, iface, cmd, NULL);
        xlog(LOG_ERR, "error: unable to exec %s: %s",
             PATH_IFCONFIG, xstrerror(errno));
        exit(1);
    }

    /* We're the parent.  Wait for the child to die. */
    if (waitpid(pid, &status, 0) == (pid_t)-1) {
        return errno;
    }

    /* Check for a happy exit code */
    if (WIFEXITED(status) && WEXITSTATUS(status) == 0) {
        return 0;
    }

    /* ifconfig failed, but we don't know why. */
    return EBADEXIT;
}


#define ST_START 0
#define ST_NUM 1
#define ST_ZERO 2
#define ST_END 3

/* Parse a PVC ATM address of te form [itf.]vpi.vci */
static int
parse_pvc(const char *const string, struct sockaddr *addr, size_t addrlen)
{
    struct sockaddr_atmpvc name;
    const char *mark = NULL;
    const char *point;
    char *p;
    long values[3];
    int state = ST_START;
    int i, ch;

    i = 0;
    for (point = string; state != ST_END; point++) {
        ch = *point;
        switch (state) {
            case ST_START:
                if (i >= 3) {
                    return EINVAL;
                }

                if (ch == '0') {
                    state = ST_ZERO;
                } else if (isdigit(ch)) {
                    mark = point;
                    state = ST_NUM;
                } else {
                    return EINVAL;
                }
                break;

            case ST_NUM:
                if (ch == '.' || ch == 0) {
                    if ((values[i++] = strtol(mark, &p, 0)) == 0) {
                        return errno;
                    }

                    if (p != point) {
                        return EINVAL;
                    }

                    state = (ch == 0) ? ST_END : ST_START;
                } else if (!isdigit(ch)) {
                    return EINVAL;
                }

                break;

            case ST_ZERO:
                if (ch != '.' && ch != 0) {
                    return EINVAL;
                }

                values[i++] = 0;
                state = ST_START;
                break;

            default:
                abort();
        }
    }

    /* Check the number of numbers found and rearrange */
    if (state != ST_END || i < 2) {
        return EINVAL;
    } else if (i == 2) {
        values[2] = values[1];
        values[1] = values[0];
        values[0] = -1;
    } else if (i > 3) {
        return EINVAL;
    }

    /* Check ranges */
    if (values[0] > SHRT_MAX || 
        values[1] > ATM_MAX_VPI ||
        values[2] > ATM_MAX_VCI) {
        return ERANGE;
    }

    /* Construct a name */
    name.sap_family = AF_ATMPVC;
    name.sap_addr.itf = values[0];
    name.sap_addr.vpi = values[1];
    name.sap_addr.vci = values[2];

    /* Copy it to the result */
    memcpy(addr, &name, addrlen < sizeof(name) ? addrlen : sizeof(name));
    return 0;
}

/* Assign virtual circuit numbers */
static int
assign_vc(int index, int encap,
          struct sockaddr *addr, socklen_t addrlen,
          int *sock_out)
{
    char buffer[32];
    struct atm_newif_br2684 req;
    struct atm_qos qos;
    struct atm_backend_br2684 back;
    int sock, err;
    int bufsize = 8192;

    /* Construct the interface name */
    snprintf(buffer, sizeof(buffer), "nas%d", index);

    /* Create a socket to talk to the ATM subsystem.  Note that when
     * this socket is closed, the bridge disappears, but the interface
     * doesn't. */
    if ((sock = socket(PF_ATMPVC, SOCK_DGRAM, ATM_AAL5)) < 0) {
        return errno;
    }

    /* Create the interface that pppd will talk to */
    memset(&req, 0, sizeof(req));
    req.backend_num = ATM_BACKEND_BR2684;
    req.media = BR2684_MEDIA_ETHERNET;
    req.mtu = 1500;
    strncpy(req.ifname, buffer, sizeof(req.ifname));

    if ((ioctl(sock, ATM_NEWBACKENDIF, &req)) != 0) {
        if (errno != EEXIST) {
            err = errno;
            close(sock);
            return err;
        }
    }

    /* Set the size of the send buffer */
    if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &bufsize, sizeof(bufsize)) < 0) {
        err = errno;
        close(sock);
        return err;
    }

    /* Set the quality of service */
    memset(&qos, 0, sizeof(qos));
    qos.aal = ATM_AAL5;
    qos.txtp.traffic_class = ATM_UBR;
    qos.txtp.max_sdu = 1524;
    qos.txtp.pcr = ATM_MAX_PCR;
    qos.rxtp.traffic_class = ATM_UBR;
    qos.rxtp.max_sdu = 1524;
    qos.rxtp.pcr = ATM_MAX_PCR;

    if (setsockopt(sock, SOL_ATM, SO_ATMQOS, &qos, sizeof(qos)) < 0) {
        err = errno;
        close(sock);
        return err;
    }

    /* Connect it to the right virtual circuit */
    if (connect(sock, addr, addrlen) < 0) {
        err = errno;
        close(sock);
        return err;
    }

    /* Configure the back-end information */
    memset(&back, 0, sizeof(back));
    back.backend_num = ATM_BACKEND_BR2684;
    back.ifspec.method = BR2684_FIND_BYIFNAME;
    strncpy(back.ifspec.spec.ifname, buffer, sizeof(back.ifspec.spec.ifname));
    back.fcs_in = BR2684_FCSIN_NO;
    back.fcs_out = BR2684_FCSOUT_NO;
    back.encaps = encap;

    if (ioctl(sock, ATM_SETBACKEND, &back) != 0) {
        err = errno;
        close(sock);
        return err;
    }

    /* Bring up the interface */
    if ((err = ifconfig(buffer, "up")) != 0) {
        close(sock);
        return err;
    }

    /* Return the file descriptor */
    if (sock_out) {
        *sock_out = sock;
    }

    return 0;
}

/* Unassign a virtual circuit */
static int
unassign_vc(int sock, int index)
{
    char buffer[32];
    int err;

    /* Construct the interface name */
    snprintf(buffer, sizeof(buffer), "nas%d", index);

    /* Bring down the interface */
    if ((err = ifconfig(buffer, "down")) != 0) {
        close(sock);
        return err;
    }

    /* Close the socket to disable the ATM back-end */
    if (close(sock) < 0) {
        return errno;
    }

    return 0;
}

/* Locates the specified device in /proc/net/atm/devices.  An itf less
 * than zero is treated as a wildcard, as is a NULL mac_addr.  If
 * neither is specified then the first device in the list will be
 * used. */
static int find_device(int itf, const char *mac_addr,
                       int *itf_out, char *path_out, size_t path_size)
{
    FILE *file;
    int iface;
    char typestr[9];
    char mac[BUFFER_SIZE];
    char buffer[BUFFER_SIZE];
    int found = 0;
    int err = 0;

    /* Open the file */
    xlog(LOG_DEBUG, "looking for interface %d in %s",
         itf, PATH_PROC_DEVICES);

    if ((file = fopen(PATH_PROC_DEVICES, "r")) == NULL) {
        return errno;
    }

    /* Skip the first line */
    if (fgets(buffer, BUFFER_SIZE, file) == NULL) {
        if (ferror(file)) {
            err = errno;
            fclose(file);
            return err;
        }
    }

    /* Scan through the devices */
    while (fscanf(file, "%3d %8s %s %1023[^\n]\n", &iface, typestr, mac, buffer) == 4) {
        /* Compare the ATM interface numbers */
        if (itf >= 0 && iface != itf) {
            continue;
        }

        /* Compare the mac address */
        if (mac_addr != NULL) {
            if (strcmp(mac_addr, mac) != 0) {
                continue;
            }
        }

        /* Otherwise we have a match */
        *itf_out = iface;
        snprintf(path_out, path_size, PATH_PROC_ATM "/%s:%d", typestr, iface);
        found = 1;
        break;
    }

    /* Check if an error occurred */
    if (ferror(file)) {
        err = errno;
    }

    /* Close the file */
    if (fclose(file) == EOF) {
        err = err ? err : errno;
    }

    if (!found) {
        err = err ? err : ENOENT;
    }
    return err;
}
                       
/* Clean up on exit */
static void clean_up_on_exit()
{
    /* Remove the pid file */
    unlink(pid_file);

    /* Kill pppd */
    if (pppd_pid != (pid_t)-1 && pppd_pid != 0) {
        kill(pppd_pid, SIGTERM);
        waitpid(pppd_pid, NULL, 0);
    }

    /* Unassign the virtual circuit */
    if (atm_fd != -1) {
        unassign_vc(atm_fd, atm_addr.sap_addr.itf);
        atm_fd = -1;
    }

    /* Unload modules */
    if (mod_list) {
        modules_remove(mod_list);
    }
}

/* Run pppd with the given args */
static int run_pppd(int argc, char **argv, pid_t *pid_out)
{
    int err;
    char buffer[BUFFER_SIZE];
    int offset = 0;
    char **args;
    int i;
    long fd_count;
    int fd;
    pid_t pid;

    /* Allocate an array for args */
    if ((args = alloca((argc + 3) * sizeof(char *))) == NULL) {
        return errno;
    }

    /* Construct pppd's argv */
    args[0] = "pppd";
    args[1] = "nodetach";
    for (i = 0; i < argc; i++) {
        args[i + 2] = argv[i];
    }
    args[argc + 2] = NULL;

    /* Copy the args into buffer for logging */
    offset = 0;
    buffer[0] = '\0';
    for (i = 0; args[i] != NULL && offset < BUFFER_SIZE; i++) {
        offset += snprintf(buffer + offset, BUFFER_SIZE - offset, "%s ", args[i]);
    }

    /* Fork a child process */
    if ((pid = fork()) == (pid_t)-1) {
        err = errno;
        free(args);
        *pid_out = pid;
        return err;
    }

    /* If we're the parent then return */
    if (pid != 0) {
        *pid_out = pid;
        xlog(LOG_DEBUG, "spawned child pid=%u", *pid_out);
        xlog(LOG_DEBUG, "argv is: %s", buffer);
        return 0;
    }

    /* We're the child.  Close off all file descriptors */
    fd_count = sysconf(_SC_OPEN_MAX);
    for (fd = 0; fd < fd_count; fd++) {
        close(fd);
    }

    /* Replace stdin, stdout and stderr with /dev/null */
    if (open(_PATH_DEVNULL, O_RDONLY) < 0 ||
        (fd = open(_PATH_DEVNULL, O_WRONLY)) < 0 ||
        dup2(fd, 2) < 0) {
        xlog(LOG_ERR, "error: unable to set up stdin, stdout and stderr: %s",
             xstrerror(errno));
        exit(EXIT_FATAL_ERROR);
    }

    /* Run pppd */
    execv(PATH_PPPD, args);
    syslog(LOG_ERR, "error: execv %s failed: %s", PATH_PPPD, xstrerror(errno));
    exit(EXIT_FATAL_ERROR);
}

int main(int argc, char *argv[])
{
    int err;
    int choice, i;
    int facility = LOG_DAEMON;
    fsm_state_t state = FSM_START;
    status_t status, last_status;
    int malformed_count = 0;
    int autherr_count = 0;
    struct timeval begin, now;
    struct sigaction action;
    int do_daemonize = 1;
    int do_pppoe = 0;
    int encap = BR2684_ENCAPS_LLC;
    int itf, found;
    long timediff;
    FILE *file;
    char *p;

    /* Determine the name of the executable */
    if ((progname = strrchr(argv[0], '/')) == NULL) {
        progname = argv[0];
    } else {
        progname++;
    }

    /* Parse the command-line options */
    while ((choice = getopt(argc, argv, OPTIONS)) != EOF) {
        switch (choice)
        {
            case 'a':
                if ((err = parse_pvc(optarg,
                                     (struct sockaddr *)&atm_addr,
                                     sizeof(atm_addr))) != 0) {
                    xlog(LOG_ERR, "error: invalid ATM address: %s", optarg);
                    exit(1);
                }

                /* Set the interface number consistently */
                if (atm_addr.sap_addr.itf < 0) {
                    atm_addr.sap_addr.itf = atm_itf;
                } else {
                    atm_itf = atm_addr.sap_addr.itf;
                }

                do_pppoe = 1;
                break;

            case 'e':
                if (strcasecmp(optarg, "vcmux") == 0) {
                    encap = BR2684_ENCAPS_VC;
                } else if (strcasecmp(optarg, "llc") == 0) {
                    encap = BR2684_ENCAPS_LLC;
                } else {
                    xlog(LOG_ERR, "error: invalid encapsulation: %s", optarg);
                    exit(1);
                }

                break;

            case 'F':
                do_daemonize = 0;
                break;

            case 'f':
                facility = -1;

                /* Look for a symbolic facility */
                for (i = 0; facilitynames[i].c_name != NULL; i++) {
                    if (strcasecmp(facilitynames[i].c_name, optarg) == 0) {
                        facility = facilitynames[i].c_val;
                        break;
                    }
                }

                /* Complain if we didn't find the facility */
                if (facility < 0) {
                    xlog(LOG_ERR, "error: unknown facility: %s", optarg);
                    exit(1);
                }

                break;

            case 'i':
                atm_itf = strtol(optarg, &p, 0);
                if (*p != 0 || atm_itf < 0) {
                    xlog(LOG_ERR, "invalid ATM interface number: %s", optarg);
                    exit(1);
                }

                if (do_pppoe) {
                    atm_addr.sap_addr.itf = atm_itf;
                }
                break;

            case 'm':
                mac_addr = optarg;
                break;

            case 'p':
                pid_file = optarg;
                break;

            case 'u':
                mod_list = optarg;
                break;

            case 'v':
                /* Look for a symbolic priority */
                found = 0;
                for (i = 0; prioritynames[i].c_name != NULL; i++) {
                    if (strcasecmp(prioritynames[i].c_name, optarg) == 0) {
                        verbosity = prioritynames[i].c_val;
                        found = 1;
                        break;
                    }
                }

                if (!found) {
                    /* Try a numeric priority */
                    verbosity = strtol(optarg, &p, 0);
                    if (*p == 0 && verbosity >= LOG_EMERG && verbosity <= LOG_DEBUG) {
                        found = 1;
                    }
                }

                /* Complain if invalid priority given */
                if (!found) {
                    verbosity = LOG_ERR;
                    xlog(LOG_ERR, "invalid verbosity level: %s", optarg);
                    exit(1);
                }

                break;

            case 'h':
                usage();
                exit(0);

            case 'V':
                fprintf(stderr, "%s (" PACKAGE ") version " VERSION "\n", progname);
                exit(0);

            case '?':
                usage();
                exit(1);

            default:
                fprintf(stderr, "%s: error: unhandled option -%c\n",
                        progname, choice);
                exit(1);
        }
    }

    /* Set up syslog */
    openlog(progname, 0, facility);

    /* Determine the path to modprobe */
    if (mod_list) {
        if ((err = find_modprobe()) != 0) {
            xlog(LOG_ERR, "error: unable to locate modprobe: %s", xstrerror(err));
            exit(1);
        }

        xlog(LOG_DEBUG, "modprobe is %s", path_modprobe);
    }

    /* We have no deceased children yet */
    morgue.pid = (pid_t)-1;

    /* Do the state machine thing */
    last_status = (status_t)-1;
    for (;;) {
        /* Shutting down? */
        if (do_shutdown) {
            /* Kill pppd */
            if (pppd_pid != (pid_t)-1 && pppd_pid != 0) {
                xlog(LOG_DEBUG, "killing pppd (pid=%d)", pppd_pid);
                if (kill(pppd_pid, SIGTERM) < 0) {
                    xlog(LOG_ERR, "error: unable to kill child process: %s",
                         xstrerror(errno));
                    exit(1);
                }

                /* Wait for it to die */
                if (waitpid(pppd_pid, NULL, 0) == (pid_t)-1) {
                    xlog(LOG_ERR, "error: waitpid failed: %s", xstrerror(errno));
                    exit(1);
                }
            }

            /* Exit */
            exit(morgue.status);
        }

        /* If we've received a signal then change state */
        if (do_restart) {
            SET_STATE(state, FSM_ERROR);
            do_restart = 0;
        }

        switch (state) {
            case FSM_START:
                if (mod_list) {
                    if ((err = modules_load(mod_list)) != 0) {
                        xlog(LOG_ERR, "modprobe %s failed: %s", mod_list, xstrerror(err));
                        exit(1);
                    }
                }

                /* Record the time now so that we can restart if the
                 * modem doesn't come up in a timely fasion. */
                xgettimeofday(&begin);

                SET_STATE(state, FSM_FIND_STATUS);
                break;

            case FSM_FIND_STATUS:
                /* Locate the device */
                if ((err = find_device(atm_itf, mac_addr,
                                       &itf, path_modem_status, sizeof(path_modem_status))) != 0) {
                    if (err != ENOENT) {
                        xlog(LOG_ERR, "error: uanble to read ATM device list: %s",
                             xstrerror(err));
                        exit(1);
                    }

                    /* See if we've waited long enough */
                    xgettimeofday(&now);
                    timediff = (now.tv_sec - begin.tv_sec) * 1000000 +
                        now.tv_usec - begin.tv_usec;

                    if (MODEM_INIT_TIMEOUT < timediff) {
                        xlog(LOG_ERR, "error: unable to locate ATM device: %s", xstrerror(err));
                        exit(1);
                    }

                    /* Otherwise sleep for a bit */
                    usleep(SHORT_PAUSE_USEC);
                    break;
                }

                /* Found it */
                xlog(LOG_INFO, "reading modem status from %s", path_modem_status);

                /* If we're doing PPPoE then record the interface
                 * number in the address */
                if (do_pppoe) {
                    atm_addr.sap_addr.itf = itf;
                }

                SET_STATE(state, FSM_FIRMWARE);
                break;

            case FSM_FIRMWARE:
                if ((err = modem_get_status(path_modem_status, &status)) != 0) {
                    if (++malformed_count < MAX_MALFORMED) {
                        xlog(LOG_WARNING, "warning: unable to determine "
                             "modem status from file %s: %s",
                             path_modem_status, xstrerror(err));
                    } else {
                        xlog(LOG_ERR, "error: unable to determine modem "
                             "status from file %s: %s",
                             path_modem_status, xstrerror(err));
                        exit(1);
                    }
                } else {
                    /* No trouble parsing the status file */
                    malformed_count = 0;

                    /* Log changes in the modem state */
                    if (last_status != status) {
                        xlog(LOG_INFO, "modem status: %s", status_names[status]);
                        last_status = status;
                    }

                    /* If the line is up then move on to the pause state */
                    if (status == STATUS_LINE_UP) {
                        SET_STATE(state, FSM_PAUSE);
                        break;
                    }
                }

                /* Is it time to give up and try again? */
                xgettimeofday(&now);

                timediff = (now.tv_sec - begin.tv_sec) * 1000000 +
                    now.tv_usec - begin.tv_usec;
                if (MODEM_LINEUP_TIMEOUT < timediff) {
                    SET_STATE(state, FSM_ERROR);
                    break;
                }

                /* Check again later */
                usleep(SHORT_PAUSE_USEC);
                break;

            case FSM_PAUSE:
                usleep(LINE_UP_SETTLE_PAUSE_USEC);
                SET_STATE(state, FSM_DAEMONIZE);
                break;

            case FSM_DAEMONIZE:
                if (do_daemonize) {
                    xlog(LOG_DEBUG, "daemonizing");

                    /* If pppd still appears to be healthy then daemonize */
                    daemonize();
                    do_daemonize = 0;
                    in_foreground = 0;

                    /* Arrange to clean up when we exit */
                    if (atexit(clean_up_on_exit) < 0) {
                        xlog(LOG_ERR, "error: atexit failed: %s", xstrerror(errno));
                        exit(1);
                    }

                    /* Write the pid to a file */
                    if ((file = fopen(pid_file, "w")) == NULL) {
                        xlog(LOG_WARNING, "warning: unable to write pid file %s: %s",
                             pid_file, xstrerror(errno));
                    } else {
                        fprintf(file, "%d\n", getpid());
                        fclose(file);
                    }
                }

                SET_STATE(state, FSM_BRIDGE);
                break;

            case FSM_BRIDGE:
                if (do_pppoe) {
                    /* Assign a virtual circuit */
                    if ((err = assign_vc(itf, encap,
                                         (struct sockaddr *)&atm_addr,
                                         sizeof(atm_addr), &atm_fd)) != 0) {
                        xlog(LOG_ERR, "%s: error: unable to assign ATM back-end: %s",
                             progname, xstrerror(err));
                        SET_STATE(state, FSM_ERROR);
                        break;
                    }
                }

                SET_STATE(state, FSM_PPPD);
                break;

            case FSM_PPPD:
                /* Set up a one-off signal handler for SIGCHLD */
                memset(&action, 0, sizeof(struct sigaction));
                action.sa_sigaction = handle_sigchld;
                action.sa_flags = SA_NOCLDSTOP | SA_RESETHAND | SA_SIGINFO;
                sigaction(SIGCHLD, &action, NULL);

                /* And one for SIGHUP */
                action.sa_sigaction = handle_sighup;
                action.sa_flags = SA_SIGINFO;
                sigaction(SIGHUP, &action, NULL);

                /* And another for SIGINT/SIGTERM */
                action.sa_sigaction = handle_sigterm;
                sigaction(SIGINT, &action, NULL);
                sigaction(SIGTERM, &action, NULL);

                /* Spawn a child process to run pppd */
                if ((err = run_pppd(argc - optind, argv + optind, &pppd_pid)) != 0) {
                    xlog(LOG_ERR, "error: unable to spawn pppd: %s", xstrerror(err));
                    exit(1);
                }

                SET_STATE(state, FSM_RUNNING);
                break;

            case FSM_RUNNING:
                /* Has pppd died on us? */
                if (morgue.pid != (pid_t)-1) {
                    /* It had better be the pppd child */
                    if (morgue.pid != pppd_pid) {
                        xlog(LOG_ERR, "internal error: dead child %d is not pppd %d",
                             morgue.pid, pppd_pid);
                        abort();
                    }

                    /* Reap the child */
                    if (waitpid(pppd_pid, NULL, 0) == (pid_t)-1) {
                        xlog(LOG_ERR, "error: waitpid failed: %s", xstrerror(errno));
                        exit(1);
                    }

                    /* The child is now dead */
                    morgue.pid = (pid_t)-1;
                    pppd_pid = (pid_t)-1;

                    /* Decide how to respond to its demise */
                    if (morgue.code == CLD_EXITED) {
                        xlog(LOG_INFO, "pppd exited with code %d", morgue.status);

                        /* Authentication failure? */
                        switch (morgue.status) {
                            case EXIT_PEER_AUTH_FAILED:
                            case EXIT_AUTH_TOPEER_FAILED:
                                if (++autherr_count >= MAX_AUTH_ERRORS) {
                                    xlog(LOG_ERR, "error: unable to authenticate with peer");
                                    exit(morgue.status);
                                }
                                break;

                            case EXIT_LOCK_FAILED:
                            case EXIT_OPEN_FAILED:
                            case EXIT_NEGOTIATION_FAILED:
                            case EXIT_PEER_DEAD:
                            case EXIT_HANGUP:
                                xlog(LOG_INFO, "pppd exited with status: %d", morgue.status);
                                break;

                            default:
                                xlog(LOG_ERR, "error: unrecoverable pppd failure: %d", morgue.status);
                                exit(morgue.status);
                        }
                    } else if (morgue.code == CLD_KILLED) {
                        xlog(LOG_INFO, "pppd exited on signal %d", morgue.status);
                    } else {
                        xlog(LOG_INFO, "pppd died with code %d, status %d",
                             morgue.code, morgue.status);
                    }

                    /* Go to the error state to recover */
                    SET_STATE(state, FSM_ERROR);
                    break;
                }

                /* Wait for a signal to wake us */
                xlog(LOG_INFO, "Running");
                pause();
                break;

            case FSM_ERROR:
                /* Kill pppd if it is still running */
                if (pppd_pid != (pid_t)-1 && pppd_pid != 0) {
                    xlog(LOG_DEBUG, "killing pppd (pid=%d)", pppd_pid);
                    if (kill(pppd_pid, SIGTERM) < 0) {
                        xlog(LOG_ERR, "error: unable to kill child process: %s",
                             xstrerror(errno));
                        exit(1);
                    }

                    /* Wait for it to die */
                    if (waitpid(pppd_pid, NULL, 0) == (pid_t)-1) {
                        xlog(LOG_ERR, "error: waitpid failed: %s", xstrerror(errno));
                        exit(1);
                    }
                }

                /* Disable the bridge interface */
                if (atm_fd != -1) {
                    err = unassign_vc(atm_fd, atm_addr.sap_addr.itf);
                    atm_fd = -1;
                    if (err != 0) {
                        xlog(LOG_ERR, "error: unable to tear down ATM back-end: %s",
                             xstrerror(err));
                    }
                }

                /* Can we power-down the modem? */
                if (mod_list != NULL) {
                    /* Should we? */
                    if ((err = modem_get_status(path_modem_status, &status)) != 0 ||
                        state != STATUS_LINE_UP) {
                        /* The ATM link is down: whack the mdem */
                        if ((err = modules_remove(mod_list)) != 0) {
                            xlog(LOG_ERR, "unable to remove modules %s: %s",
                                 mod_list, xstrerror(err));
                            exit(1);
                        }

                        /* Give the line a chance to settle before we
                         * reload the modules */
                        if (mod_list != NULL) {
                            usleep(UNLOAD_SETTLE_PAUSE_USEC);
                        }
                    }
                }

                /* Go to the start state */
                SET_STATE(state, FSM_START);
                break;

            default:
                xlog(LOG_ERR, "error: unknown state: %d", state);
                exit(1);
        }
    }

    exit(0);
}
